@@ -21,17 +21,20 @@ import ContainerResource
 import ContainerSandboxServiceClient
 import ContainerXPC
 import Containerization
+import ContainerizationEXT4
 import ContainerizationError
 import ContainerizationExtras
 import ContainerizationOCI
 import ContainerizationOS
 import Foundation
 import Logging
+import SystemPackage
 
 public actor ContainersService {
     struct ContainerState {
         var snapshot: ContainerSnapshot
         var client: SandboxClient?
+        var allocatedAttachments: [AllocatedAttachment]
 
         func getClient() throws -> SandboxClient {
             guard let client else {
@@ -49,25 +52,40 @@ public actor ContainersService {
     private static let launchdDomainString = try! ServiceManager.getDomainString()
 
     private let log: Logger
+    private let debugHelpers: Bool
     private let containerRoot: URL
     private let pluginLoader: PluginLoader
     private let runtimePlugins: [Plugin]
     private let exitMonitor: ExitMonitor
 
-    private let lock = AsyncLock()
+    private let lock: AsyncLock
     private var containers: [String: ContainerState]
 
-    public init(appRoot: URL, pluginLoader: PluginLoader, log: Logger) throws {
+    // FIXME: Find a better mechanism for services running on the APIServer to work with each other
+    private weak var networksService: NetworksService?
+
+    public init(
+        appRoot: URL,
+        pluginLoader: PluginLoader,
+        log: Logger,
+        debugHelpers: Bool = false
+    ) throws {
         let containerRoot = appRoot.appendingPathComponent("containers")
         try FileManager.default.createDirectory(at: containerRoot, withIntermediateDirectories: true)
         self.exitMonitor = ExitMonitor(log: log)
+        self.lock = AsyncLock(log: log)
         self.containerRoot = containerRoot
         self.pluginLoader = pluginLoader
         self.log = log
+        self.debugHelpers = debugHelpers
         self.runtimePlugins = pluginLoader.findPlugins().filter { $0.hasType(.runtime) }
         self.containers = try Self.loadAtBoot(root: containerRoot, loader: pluginLoader, log: log)
     }
 
+    public func setNetworksService(_ service: NetworksService) async {
+        self.networksService = service
+    }
+
     static func loadAtBoot(root: URL, loader: PluginLoader, log: Logger) throws -> [String: ContainerState] {
         var directories = try FileManager.default.contentsOfDirectory(
             at: root,
@@ -81,15 +99,16 @@ public actor ContainersService {
         var results = [String: ContainerState]()
         for dir in directories {
             do {
-                let bundle = ContainerResource.Bundle(path: dir)
-                let config = try bundle.configuration
+                let config = try Self.getContainerConfiguration(at: dir)
+
                 let state = ContainerState(
                     snapshot: .init(
                         configuration: config,
                         status: .stopped,
                         networks: [],
                         startedDate: nil
-                    )
+                    ),
+                    allocatedAttachments: []
                 )
                 results[config.id] = state
                 guard runtimePlugins.first(where: { $0.name == config.runtimeHandler }) != nil else {
@@ -100,22 +119,66 @@ public actor ContainersService {
                 }
             } catch {
                 try? FileManager.default.removeItem(at: dir)
-                log.warning("failed to load container bundle at \(dir.path)")
+                log.warning(
+                    "failed to load container",
+                    metadata: [
+                        "path": "\(dir.path)",
+                        "error": "\(error)",
+                    ])
             }
         }
         return results
     }
 
-    /// List all containers registered with the service.
-    public func list() async throws -> [ContainerSnapshot] {
-        self.log.debug("\(#function)")
-        return self.containers.values.map { $0.snapshot }
+    /// List containers matching the given filters.
+    public func list(filters: ContainerListFilters = .all) async throws -> [ContainerSnapshot] {
+        log.debug(
+            "ContainersService: enter",
+            metadata: [
+                "func": "\(#function)"
+            ]
+        )
+        defer {
+            log.debug(
+                "ContainersService: exit",
+                metadata: [
+                    "func": "\(#function)"
+                ]
+            )
+        }
+
+        return self.containers.values.compactMap { state -> ContainerSnapshot? in
+            let snapshot = state.snapshot
+
+            if !filters.ids.isEmpty {
+                guard filters.ids.contains(snapshot.id) else {
+                    return nil
+                }
+            }
+
+            if let status = filters.status {
+                guard snapshot.status == status else {
+                    return nil
+                }
+            }
+
+            for (key, value) in filters.labels {
+                guard snapshot.configuration.labels[key] == value else {
+                    return nil
+                }
+            }
+
+            return snapshot
+        }
     }
 
     /// Execute an operation with the current container list while maintaining atomicity
     /// This prevents race conditions where containers are created during the operation
-    public func withContainerList<T: Sendable>(_ operation: @Sendable @escaping ([ContainerSnapshot]) async throws -> T) async throws -> T {
-        try await lock.withLock { context in
+    public func withContainerList<T: Sendable>(
+        logMetadata: Logger.Metadata? = nil,
+        _ operation: @Sendable @escaping ([ContainerSnapshot]) async throws -> T
+    ) async throws -> T {
+        try await lock.withLock(logMetadata: logMetadata) { context in
             let snapshots = await self.containers.values.map { $0.snapshot }
             return try await operation(snapshots)
         }
@@ -124,7 +187,7 @@ public actor ContainersService {
     /// Calculate disk usage for containers
     /// - Returns: Tuple of (total count, active count, total size, reclaimable size)
     public func calculateDiskUsage() async -> (Int, Int, UInt64, UInt64) {
-        await lock.withLock { _ in
+        await lock.withLock(logMetadata: ["acquirer": "\(#function)"]) { _ in
             var totalSize: UInt64 = 0
             var reclaimableSize: UInt64 = 0
             var activeCount = 0
@@ -149,7 +212,7 @@ public actor ContainersService {
     /// Get set of image references used by containers (for disk usage calculation)
     /// - Returns: Set of image references currently in use
     public func getActiveImageReferences() async -> Set<String> {
-        await lock.withLock { _ in
+        await lock.withLock(logMetadata: ["acquirer": "\(#function)"]) { _ in
             var imageRefs = Set<String>()
             for (_, state) in await self.containers {
                 imageRefs.insert(state.snapshot.configuration.image.reference)
@@ -192,10 +255,25 @@ public actor ContainersService {
     }
 
     /// Create a new container from the provided id and configuration.
-    public func create(configuration: ContainerConfiguration, kernel: Kernel, options: ContainerCreateOptions) async throws {
-        self.log.debug("\(#function)")
+    public func create(configuration: ContainerConfiguration, kernel: Kernel, options: ContainerCreateOptions, initImage: String? = nil) async throws {
+        log.debug(
+            "ContainersService: enter",
+            metadata: [
+                "func": "\(#function)",
+                "id": "\(configuration.id)",
+            ]
+        )
+        defer {
+            log.debug(
+                "ContainersService: exit",
+                metadata: [
+                    "func": "\(#function)",
+                    "id": "\(configuration.id)",
+                ]
+            )
+        }
 
-        try await self.lock.withLock { context in
+        try await self.lock.withLock(logMetadata: ["acquirer": "\(#function)", "id": "\(configuration.id)"]) { context in
             guard await self.containers[configuration.id] == nil else {
                 throw ContainerizationError(
                     .exists,
@@ -231,21 +309,60 @@ public actor ContainersService {
                 )
             }
 
+            // Protect against a user providing a memory amount that will cause us to not be able
+            // to boot. We can go lower, but this is a somewhat safe threshold. Containerization
+            // also gives a little bit extra than the user asked for to account for guest agent overhead.
+            //
+            // NOTE: We could potentially leave this validation to the sandbox service(s), as
+            // it's possible there could be an implementation that can get away with a lower
+            // amount and be perfectly safe.
+            let minimumMemory: UInt64 = 200.mib()
+            guard configuration.resources.memoryInBytes >= minimumMemory else {
+                throw ContainerizationError(
+                    .invalidArgument,
+                    message: "minimum memory amount allowed is 200 MiB (got \(configuration.resources.memoryInBytes) bytes)"
+                )
+            }
+
             let path = self.containerRoot.appendingPathComponent(configuration.id)
             let systemPlatform = kernel.platform
-            let initFs = try await self.getInitBlock(for: systemPlatform.ociPlatform())
 
-            let bundle = try ContainerResource.Bundle.create(
-                path: path,
-                initialFilesystem: initFs,
-                kernel: kernel,
-                containerConfiguration: configuration
+            // Fetch init image (custom or default)
+            self.log.debug(
+                "ContainersService: get init block",
+                metadata: [
+                    "id": "\(configuration.id)"
+                ]
             )
+            let initFilesystem = try await self.getInitBlock(for: systemPlatform.ociPlatform(), imageRef: initImage)
+
             do {
+                self.log.debug(
+                    "create snapshot",
+                    metadata: [
+                        "id": "\(configuration.id)",
+                        "ref": "\(configuration.image.reference)",
+                    ])
                 let containerImage = ClientImage(description: configuration.image)
                 let imageFs = try await containerImage.getCreateSnapshot(platform: configuration.platform)
-                try bundle.setContainerRootFs(cloning: imageFs, readonly: configuration.readOnly)
-                try bundle.write(filename: "options.json", value: options)
+
+                self.log.debug(
+                    "configure runtime",
+                    metadata: [
+                        "id": "\(configuration.id)",
+                        "kernel": "\(kernel.path)",
+                        "initfs": "\(initImage ?? ClientImage.initImageRef)",
+                    ])
+                let runtimeConfig = RuntimeConfiguration(
+                    path: path,
+                    initialFilesystem: initFilesystem,
+                    kernel: kernel,
+                    containerConfiguration: configuration,
+                    containerRootFilesystem: imageFs,
+                    options: options
+                )
+
+                try runtimeConfig.writeRuntimeConfiguration()
 
                 let snapshot = ContainerSnapshot(
                     configuration: configuration,
@@ -253,13 +370,8 @@ public actor ContainersService {
                     networks: [],
                     startedDate: nil
                 )
-                await self.setContainerState(configuration.id, ContainerState(snapshot: snapshot), context: context)
+                await self.setContainerState(configuration.id, ContainerState(snapshot: snapshot, allocatedAttachments: []), context: context)
             } catch {
-                do {
-                    try bundle.delete()
-                } catch {
-                    self.log.error("failed to delete bundle for container \(configuration.id): \(error)")
-                }
                 throw error
             }
         }
@@ -267,8 +379,24 @@ public actor ContainersService {
 
     /// Bootstrap the init process of the container.
     public func bootstrap(id: String, stdio: [FileHandle?]) async throws {
-        self.log.debug("\(#function)")
-        try await self.lock.withLock { context in
+        log.debug(
+            "ContainersService: enter",
+            metadata: [
+                "func": "\(#function)",
+                "id": "\(id)",
+            ]
+        )
+        defer {
+            log.debug(
+                "ContainersService: exit",
+                metadata: [
+                    "func": "\(#function)",
+                    "id": "\(id)",
+                ]
+            )
+        }
+
+        try await self.lock.withLock(logMetadata: ["acquirer": "\(#function)", "id": "\(id)"]) { context in
             var state = try await self.getContainerState(id: id, context: context)
 
             // We've already bootstrapped this container. Ideally we should be able to
@@ -279,15 +407,28 @@ public actor ContainersService {
             }
 
             let path = self.containerRoot.appendingPathComponent(id)
-            let bundle = ContainerResource.Bundle(path: path)
-            let config = try bundle.configuration
+            let config = try Self.getContainerConfiguration(at: path)
 
+            var allocatedAttachments = [AllocatedAttachment]()
             do {
+                for n in config.networks {
+                    let allocatedAttach = try await self.networksService?.allocate(
+                        id: n.network,
+                        hostname: n.options.hostname,
+                        macAddress: n.options.macAddress
+                    )
+                    guard let allocatedAttach = allocatedAttach else {
+                        throw ContainerizationError(.internalError, message: "failed to allocate a network")
+                    }
+                    allocatedAttachments.append(allocatedAttach)
+                }
+
                 try Self.registerService(
                     plugin: self.runtimePlugins.first { $0.name == config.runtimeHandler }!,
                     loader: self.pluginLoader,
                     configuration: config,
-                    path: path
+                    path: path,
+                    debug: self.debugHelpers
                 )
 
                 let runtime = state.snapshot.configuration.runtimeHandler
@@ -295,7 +436,7 @@ public actor ContainersService {
                     id: id,
                     runtime: runtime
                 )
-                try await sandboxClient.bootstrap(stdio: stdio)
+                try await sandboxClient.bootstrap(stdio: stdio, allocatedAttachments: allocatedAttachments)
 
                 try await self.exitMonitor.registerProcess(
                     id: id,
@@ -303,8 +444,23 @@ public actor ContainersService {
                 )
 
                 state.client = sandboxClient
+                state.allocatedAttachments = allocatedAttachments
                 await self.setContainerState(id, state, context: context)
             } catch {
+                for allocatedAttach in allocatedAttachments {
+                    do {
+                        try await self.networksService?.deallocate(attachment: allocatedAttach.attachment)
+                    } catch {
+                        self.log.error(
+                            "failed to deallocate network attachment",
+                            metadata: [
+                                "id": "\(id)",
+                                "network": "\(allocatedAttach.attachment.network)",
+                                "error": "\(error)",
+                            ])
+                    }
+                }
+
                 let label = Self.fullLaunchdServiceLabel(
                     runtimeName: config.runtimeHandler,
                     instanceId: id
@@ -312,7 +468,6 @@ public actor ContainersService {
 
                 await self.exitMonitor.stopTracking(id: id)
                 try? ServiceManager.deregister(fullServiceLabel: label)
-
                 throw error
             }
         }
@@ -325,7 +480,24 @@ public actor ContainersService {
         config: ProcessConfiguration,
         stdio: [FileHandle?]
     ) async throws {
-        self.log.debug("\(#function)")
+        log.debug(
+            "ContainersService: enter",
+            metadata: [
+                "func": "\(#function)",
+                "id": "\(id)",
+                "processId": "\(processID)",
+                "command": "\(config.arguments.isEmpty ? "" : config.arguments[0])",
+            ]
+        )
+        defer {
+            log.debug(
+                "ContainersService: exit",
+                metadata: [
+                    "func": "\(#function)",
+                    "id": "\(id)",
+                ]
+            )
+        }
 
         let state = try self._getContainerState(id: id)
         let client = try state.getClient()
@@ -340,9 +512,26 @@ public actor ContainersService {
     /// createProcess, or the init process of the container which requires
     /// id == processID.
     public func startProcess(id: String, processID: String) async throws {
-        self.log.debug("\(#function)")
+        log.debug(
+            "ContainersService: enter",
+            metadata: [
+                "func": "\(#function)",
+                "id": "\(id)",
+                "processId": "\(processID)",
+            ]
+        )
+        defer {
+            log.debug(
+                "ContainersService: exit",
+                metadata: [
+                    "func": "\(#function)",
+                    "id": "\(id)",
+                    "processId": "\(processID)",
+                ]
+            )
+        }
 
-        try await self.lock.withLock { context in
+        try await self.lock.withLock(logMetadata: ["acquirer": "\(#function)", "id": "\(id)", "processId": "\(processID)"]) { context in
             var state = try await self.getContainerState(id: id, context: context)
 
             let isInit = Self.isInitProcess(id: id, processID: processID)
@@ -360,9 +549,14 @@ public actor ContainersService {
             do {
                 let log = self.log
                 let waitFunc: ExitMonitor.WaitHandler = {
-                    log.info("registering container \(id) with exit monitor")
+                    log.info("registering container with exit monitor")
                     let code = try await client.wait(id)
-                    log.info("container \(id) finished in exit monitor, exit code \(code)")
+                    log.info(
+                        "container finished in exit monitor",
+                        metadata: [
+                            "id": "\(id)",
+                            "rc": "\(code)",
+                        ])
 
                     return code
                 }
@@ -376,7 +570,6 @@ public actor ContainersService {
             } catch {
                 await self.exitMonitor.stopTracking(id: id)
                 try? await client.stop(options: ContainerStopOptions.default)
-
                 throw error
             }
         }
@@ -384,7 +577,25 @@ public actor ContainersService {
 
     /// Send a signal to the container.
     public func kill(id: String, processID: String, signal: Int64) async throws {
-        self.log.debug("\(#function)")
+        log.debug(
+            "ContainersService: enter",
+            metadata: [
+                "func": "\(#function)",
+                "id": "\(id)",
+                "processId": "\(processID)",
+                "signal": "\(signal)",
+            ]
+        )
+        defer {
+            log.debug(
+                "ContainersService: exit",
+                metadata: [
+                    "func": "\(#function)",
+                    "id": "\(id)",
+                    "processId": "\(processID)",
+                ]
+            )
+        }
 
         let state = try self._getContainerState(id: id)
         let client = try state.getClient()
@@ -394,7 +605,22 @@ public actor ContainersService {
     /// Stop all containers inside the sandbox, aborting any processes currently
     /// executing inside the container, before stopping the underlying sandbox.
     public func stop(id: String, options: ContainerStopOptions) async throws {
-        self.log.debug("\(#function)")
+        log.debug(
+            "ContainersService: enter",
+            metadata: [
+                "func": "\(#function)",
+                "id": "\(id)",
+            ]
+        )
+        defer {
+            log.debug(
+                "ContainersService: exit",
+                metadata: [
+                    "func": "\(#function)",
+                    "id": "\(id)",
+                ]
+            )
+        }
 
         let state = try self._getContainerState(id: id)
 
@@ -417,6 +643,25 @@ public actor ContainersService {
     }
 
     public func dial(id: String, port: UInt32) async throws -> FileHandle {
+        log.debug(
+            "ContainersService: enter",
+            metadata: [
+                "func": "\(#function)",
+                "id": "\(id)",
+                "port": "\(port)",
+            ]
+        )
+        defer {
+            log.debug(
+                "ContainersService: exit",
+                metadata: [
+                    "func": "\(#function)",
+                    "id": "\(id)",
+                    "port": "\(port)",
+                ]
+            )
+        }
+
         let state = try self._getContainerState(id: id)
         let client = try state.getClient()
         return try await client.dial(port)
@@ -425,7 +670,24 @@ public actor ContainersService {
     /// Wait waits for the container's init process or exec to exit and returns the
     /// exit status.
     public func wait(id: String, processID: String) async throws -> ExitStatus {
-        self.log.debug("\(#function)")
+        log.debug(
+            "ContainersService: enter",
+            metadata: [
+                "func": "\(#function)",
+                "id": "\(id)",
+                "processId": "\(processID)",
+            ]
+        )
+        defer {
+            log.debug(
+                "ContainersService: exit",
+                metadata: [
+                    "func": "\(#function)",
+                    "id": "\(id)",
+                    "processId": "\(processID)",
+                ]
+            )
+        }
 
         let state = try self._getContainerState(id: id)
         let client = try state.getClient()
@@ -434,7 +696,24 @@ public actor ContainersService {
 
     /// Resize resizes the container's PTY if one exists.
     public func resize(id: String, processID: String, size: Terminal.Size) async throws {
-        self.log.debug("\(#function)")
+        log.trace(
+            "ContainersService: enter",
+            metadata: [
+                "func": "\(#function)",
+                "id": "\(id)",
+                "processId": "\(processID)",
+            ]
+        )
+        defer {
+            log.trace(
+                "ContainersService: exit",
+                metadata: [
+                    "func": "\(#function)",
+                    "id": "\(id)",
+                    "processId": "\(processID)",
+                ]
+            )
+        }
 
         let state = try self._getContainerState(id: id)
         let client = try state.getClient()
@@ -443,11 +722,29 @@ public actor ContainersService {
 
     // Get the logs for the container.
     public func logs(id: String) async throws -> [FileHandle] {
-        self.log.debug("\(#function)")
+        log.debug(
+            "ContainersService: enter",
+            metadata: [
+                "func": "\(#function)",
+                "id": "\(id)",
+            ]
+        )
+        defer {
+            log.debug(
+                "ContainersService: exit",
+                metadata: [
+                    "func": "\(#function)",
+                    "id": "\(id)",
+                ]
+            )
+        }
 
         // Logs doesn't care if the container is running or not, just that
-        // the bundle is there, and that the files actually exist.
+        // the bundle is there, and that the files actually exist. We do
+        // first try and get the container state so we get a nicer error message
+        // (container foo not found) however.
         do {
+            _ = try _getContainerState(id: id)
             let path = self.containerRoot.appendingPathComponent(id)
             let bundle = ContainerResource.Bundle(path: path)
             return [
@@ -464,7 +761,22 @@ public actor ContainersService {
 
     /// Get statistics for the container.
     public func stats(id: String) async throws -> ContainerStats {
-        self.log.debug("\(#function)")
+        log.debug(
+            "ContainersService: enter",
+            metadata: [
+                "func": "\(#function)",
+                "id": "\(id)",
+            ]
+        )
+        defer {
+            log.debug(
+                "ContainersService: exit",
+                metadata: [
+                    "func": "\(#function)",
+                    "id": "\(id)",
+                ]
+            )
+        }
 
         let state = try self._getContainerState(id: id)
         let client = try state.getClient()
@@ -473,7 +785,24 @@ public actor ContainersService {
 
     /// Delete a container and its resources.
     public func delete(id: String, force: Bool) async throws {
-        self.log.debug("\(#function)")
+        log.info(
+            "ContainersService: enter",
+            metadata: [
+                "func": "\(#function)",
+                "id": "\(id)",
+                "force": "\(force)",
+            ]
+        )
+        defer {
+            log.debug(
+                "ContainersService: exit",
+                metadata: [
+                    "func": "\(#function)",
+                    "id": "\(id)",
+                ]
+            )
+        }
+
         let state = try self._getContainerState(id: id)
         switch state.snapshot.status {
         case .running:
@@ -489,8 +818,22 @@ public actor ContainersService {
             )
             let client = try state.getClient()
             try await client.stop(options: opts)
-            try await self.lock.withLock { context in
-                try await self.cleanup(id: id, context: context)
+            try await self.lock.withLock(logMetadata: ["acquirer": "\(#function)", "id": "\(id)"]) { context in
+                self.log.info(
+                    "ContainersService: attempt cleanup",
+                    metadata: [
+                        "func": "\(#function)",
+                        "id": "\(id)",
+                    ]
+                )
+                try await self.cleanUp(id: id, context: context)
+                self.log.info(
+                    "ContainersService: successful cleanup",
+                    metadata: [
+                        "func": "\(#function)",
+                        "id": "\(id)",
+                    ]
+                )
             }
         case .stopping:
             throw ContainerizationError(
@@ -498,29 +841,63 @@ public actor ContainersService {
                 message: "container \(id) is \(state.snapshot.status) and can not be deleted"
             )
         default:
-            try await self.lock.withLock { context in
-                try await self.cleanup(id: id, context: context)
+            try await self.lock.withLock(logMetadata: ["acquirer": "\(#function)", "id": "\(id)"]) { context in
+                try await self.cleanUp(id: id, context: context)
             }
         }
     }
 
     public func containerDiskUsage(id: String) async throws -> UInt64 {
-        self.log.debug("\(#function)")
+        log.debug(
+            "ContainersService: enter",
+            metadata: [
+                "func": "\(#function)",
+                "id": "\(id)",
+            ]
+        )
+        defer {
+            log.debug(
+                "ContainersService: exit",
+                metadata: [
+                    "func": "\(#function)",
+                    "id": "\(id)",
+                ]
+            )
+        }
 
         let containerPath = self.containerRoot.appendingPathComponent(id).path
 
         return Self.calculateDirectorySize(at: containerPath)
     }
 
+    public func exportRootfs(id: String, archive: URL) async throws {
+        self.log.debug("\(#function)")
+
+        let state = try self._getContainerState(id: id)
+        guard state.snapshot.status == .stopped else {
+            throw ContainerizationError(.invalidState, message: "container is not stopped")
+        }
+
+        let path = self.containerRoot.appendingPathComponent(id)
+        let bundle = ContainerResource.Bundle(path: path)
+        let rootfs = bundle.containerRootfsBlock
+        try EXT4.EXT4Reader(blockDevice: FilePath(rootfs)).export(archive: FilePath(archive))
+    }
+
     private func handleContainerExit(id: String, code: ExitStatus? = nil) async throws {
-        try await self.lock.withLock { [self] context in
+        try await self.lock.withLock(logMetadata: ["acquirer": "\(#function)", "id": "\(id)"]) { [self] context in
             try await handleContainerExit(id: id, code: code, context: context)
         }
     }
 
     private func handleContainerExit(id: String, code: ExitStatus?, context: AsyncLock.Context) async throws {
         if let code {
-            self.log.info("Handling container \(id) exit. Code \(code)")
+            self.log.info(
+                "handling container exit",
+                metadata: [
+                    "id": "\(id)",
+                    "rc": "\(code)",
+                ])
         }
 
         var state: ContainerState
@@ -537,7 +914,7 @@ public actor ContainersService {
         await self.exitMonitor.stopTracking(id: id)
 
         // Shutdown and deregister the sandbox service
-        self.log.info("Shutting down sandbox service for \(id)")
+        self.log.info("shutting down sandbox service", metadata: ["id": "\(id)"])
 
         let path = self.containerRoot.appendingPathComponent(id)
         let bundle = ContainerResource.Bundle(path: path)
@@ -554,7 +931,12 @@ public actor ContainersService {
             do {
                 try await client.shutdown()
             } catch {
-                self.log.error("Failed to shutdown sandbox service for \(id): \(error)")
+                self.log.error(
+                    "failed to shutdown sandbox service",
+                    metadata: [
+                        "id": "\(id)",
+                        "error": "\(error)",
+                    ])
             }
         }
 
@@ -563,19 +945,42 @@ public actor ContainersService {
         // the process was killed externally.
         do {
             try ServiceManager.deregister(fullServiceLabel: label)
-            self.log.info("Deregistered sandbox service for \(id)")
+            self.log.info("deregistered sandbox service", metadata: ["id": "\(id)"])
         } catch {
-            self.log.error("Failed to deregister sandbox service for \(id): \(error)")
+            self.log.error(
+                "failed to deregister sandbox service",
+                metadata: [
+                    "id": "\(id)",
+                    "error": "\(error)",
+                ])
+        }
+
+        // Best effort deallocate network attachments for the container. Don't throw on
+        // failure so we can continue with state cleanup.
+        self.log.info("deallocating network attachments", metadata: ["id": "\(id)"])
+        for allocatedAttach in state.allocatedAttachments {
+            do {
+                try await self.networksService?.deallocate(attachment: allocatedAttach.attachment)
+            } catch {
+                self.log.error(
+                    "failed to deallocate network attachment",
+                    metadata: [
+                        "id": "\(id)",
+                        "network": "\(allocatedAttach.attachment.network)",
+                        "error": "\(error)",
+                    ])
+            }
         }
 
         state.snapshot.status = .stopped
         state.snapshot.networks = []
         state.client = nil
+        state.allocatedAttachments = []
         await self.setContainerState(id, state, context: context)
 
         let options = try getContainerCreationOptions(id: id)
         if options.autoRemove {
-            try await self.cleanup(id: id, context: context)
+            try await self.cleanUp(id: id, context: context)
         }
     }
 
@@ -583,8 +988,23 @@ public actor ContainersService {
         "\(Self.launchdDomainString)/\(Self.machServicePrefix).\(runtimeName).\(instanceId)"
     }
 
-    private func _cleanup(id: String) async throws {
-        self.log.debug("\(#function)")
+    private func _cleanUp(id: String) async throws {
+        log.debug(
+            "ContainersService: enter",
+            metadata: [
+                "func": "\(#function)",
+                "id": "\(id)",
+            ]
+        )
+        defer {
+            log.debug(
+                "ContainersService: exit",
+                metadata: [
+                    "func": "\(#function)",
+                    "id": "\(id)",
+                ]
+            )
+        }
 
         // Did the exit container handler win?
         if self.containers[id] == nil {
@@ -596,20 +1016,50 @@ public actor ContainersService {
         // the OCI runtime.
         await self.exitMonitor.stopTracking(id: id)
         let path = self.containerRoot.appendingPathComponent(id)
-        let bundle = ContainerResource.Bundle(path: path)
-        let config = try bundle.configuration
 
-        let label = Self.fullLaunchdServiceLabel(
-            runtimeName: config.runtimeHandler,
-            instanceId: id
-        )
-        try ServiceManager.deregister(fullServiceLabel: label)
-        try bundle.delete()
+        // Try to get config for service deregistration
+        // Don't fail if bundle is incomplete
+        var config: ContainerConfiguration?
+        let bundle = ContainerResource.Bundle(path: path)
+        do {
+            config = try bundle.configuration
+        } catch {
+            self.log.warning(
+                "failed to read bundle configuration during cleanup for container",
+                metadata: [
+                    "id": "\(id)",
+                    "error": "\(error)",
+                ])
+        }
+
+        // Only try to deregister service if we have a valid config
+        // TODO: Change this so we don't have to reread the config
+        // possibly store the container ID to service label mapping
+        if let config = config {
+            let label = Self.fullLaunchdServiceLabel(
+                runtimeName: config.runtimeHandler,
+                instanceId: id
+            )
+            try? ServiceManager.deregister(fullServiceLabel: label)
+        }
+
+        // Always try to delete the bundle directory, even if it's incomplete
+        do {
+            try bundle.delete()
+        } catch {
+            self.log.warning(
+                "failed to delete bundle for container",
+                metadata: [
+                    "id": "\(id)",
+                    "error": "\(error)",
+                ])
+        }
+
         self.containers.removeValue(forKey: id)
     }
 
-    private func cleanup(id: String, context: AsyncLock.Context) async throws {
-        try await self._cleanup(id: id)
+    private func cleanUp(id: String, context: AsyncLock.Context) async throws {
+        try await self._cleanUp(id: id)
     }
 
     private func getContainerCreationOptions(id: String) throws -> ContainerCreateOptions {
@@ -619,8 +1069,9 @@ public actor ContainersService {
         return options
     }
 
-    private func getInitBlock(for platform: Platform) async throws -> Filesystem {
-        let initImage = try await ClientImage.fetch(reference: ClientImage.initImageRef, platform: platform)
+    private func getInitBlock(for platform: Platform, imageRef: String? = nil) async throws -> Filesystem {
+        let ref = imageRef ?? ClientImage.initImageRef
+        let initImage = try await ClientImage.fetch(reference: ref, platform: platform)
         var fs = try await initImage.getCreateSnapshot(platform: platform)
         fs.options = ["ro"]
         return fs
@@ -630,14 +1081,15 @@ public actor ContainersService {
         plugin: Plugin,
         loader: PluginLoader,
         configuration: ContainerConfiguration,
-        path: URL
+        path: URL,
+        debug: Bool
     ) throws {
         let args = [
             "start",
             "--root", path.path,
             "--uuid", configuration.id,
-            "--debug",
-        ]
+            debug ? "--debug" : nil,
+        ].compactMap { $0 }
         try loader.registerWithLaunchd(
             plugin: plugin,
             pluginStateRoot: path,
@@ -668,6 +1120,22 @@ public actor ContainersService {
     private static func isInitProcess(id: String, processID: String) -> Bool {
         id == processID
     }
+
+    /// Get container configuration, either from existing bundle or from RuntimeConfiguration
+    private static func getContainerConfiguration(at path: URL) throws -> ContainerConfiguration {
+        let bundle = ContainerResource.Bundle(path: path)
+        do {
+            return try bundle.configuration
+        } catch {
+            // Bundle doesn't exist or incomplete, try runtime configuration
+            // This handles containers that were created but not started yet
+            let runtimeConfig = try RuntimeConfiguration.readRuntimeConfiguration(from: path)
+            guard let config = runtimeConfig.containerConfiguration else {
+                throw ContainerizationError(.internalError, message: "runtime configuration missing container configuration")
+            }
+            return config
+        }
+    }
 }
 
 extension XPCMessage {
