@@ -42,4 +42,10 @@ public enum SandboxKeys: String {
 
     /// Container statistics
     case statistics
+
+    /// Network resource keys.
+    case allocatedAttachments
+    case networkAdditionalData
+    case networkAttachment
+    case networkPluginInfo
 }
