@@ -329,20 +329,6 @@ struct ParserTest {
 
     @Test
     func testRelativePaths() throws {
-        let originalDir = FileManager.default.currentDirectoryPath
-        defer {
-            FileManager.default.changeCurrentDirectoryPath(originalDir)
-        }
-
-        func realpath(_ path: String) -> String {
-            let resolved = UnsafeMutablePointer<CChar>.allocate(capacity: Int(PATH_MAX))
-            defer { resolved.deallocate() }
-            guard let result = Darwin.realpath(path, resolved) else {
-                return path
-            }
-            return String(cString: result)
-        }
-
         // Test bind mount with relative path "."
         do {
             let tempDir = FileManager.default.temporaryDirectory.appendingPathComponent("test-bind-\(UUID().uuidString)")
@@ -351,12 +337,11 @@ struct ParserTest {
                 try? FileManager.default.removeItem(at: tempDir)
             }
 
-            FileManager.default.changeCurrentDirectoryPath(tempDir.path)
-            let result = try Parser.mount("type=bind,src=.,dst=/foo")
+            let result = try Parser.mount("type=bind,src=.,dst=/foo", relativeTo: tempDir)
 
             switch result {
             case .filesystem(let fs):
-                #expect(fs.source == realpath(tempDir.path))
+                #expect(fs.source == tempDir.standardizedFileURL.path)
                 #expect(fs.destination == "/foo")
                 #expect(!fs.isVolume)
             case .volume:
@@ -372,12 +357,11 @@ struct ParserTest {
                 try? FileManager.default.removeItem(at: tempDir)
             }
 
-            FileManager.default.changeCurrentDirectoryPath(tempDir.path)
-            let result = try Parser.volume("./:/foo")
+            let result = try Parser.volume("./:/foo", relativeTo: tempDir)
 
             switch result {
             case .filesystem(let fs):
-                let expectedPath = realpath(tempDir.path)
+                let expectedPath = tempDir.standardizedFileURL.path
                 // Normalize trailing slashes for comparison
                 #expect(fs.source.trimmingCharacters(in: CharacterSet(charactersIn: "/")) == expectedPath.trimmingCharacters(in: CharacterSet(charactersIn: "/")))
                 #expect(fs.destination == "/foo")
@@ -395,12 +379,11 @@ struct ParserTest {
                 try? FileManager.default.removeItem(at: tempDir)
             }
 
-            FileManager.default.changeCurrentDirectoryPath(tempDir.path)
-            let result = try Parser.volume("./subdir:/foo")
+            let result = try Parser.volume("./subdir:/foo", relativeTo: tempDir)
 
             switch result {
             case .filesystem(let fs):
-                let expectedPath = realpath(nestedDir.path)
+                let expectedPath = nestedDir.standardizedFileURL.path
                 // Normalize trailing slashes for comparison
                 #expect(fs.source.trimmingCharacters(in: CharacterSet(charactersIn: "/")) == expectedPath.trimmingCharacters(in: CharacterSet(charactersIn: "/")))
                 #expect(fs.destination == "/foo")
@@ -570,6 +553,42 @@ struct ParserTest {
         #expect(result == ["EMPTY="])
     }
 
+    @Test
+    func testAllEnvUserOverridesImage() throws {
+        let result = try Parser.allEnv(
+            imageEnvs: ["FOO=fromimage", "BAR=kept"],
+            envFiles: [],
+            envs: ["FOO=fromuser"]
+        )
+        #expect(Set(result) == Set(["FOO=fromuser", "BAR=kept"]))
+    }
+
+    @Test
+    func testAllEnvFileOverridesImage() throws {
+        let tmpFile = try tmpFileWithContent("FOO=fromfile\n")
+        defer { try? FileManager.default.removeItem(at: tmpFile) }
+
+        let result = try Parser.allEnv(
+            imageEnvs: ["FOO=fromimage", "BAR=kept"],
+            envFiles: [tmpFile.path],
+            envs: []
+        )
+        #expect(Set(result) == Set(["FOO=fromfile", "BAR=kept"]))
+    }
+
+    @Test
+    func testAllEnvUserOverridesFileOverridesImage() throws {
+        let tmpFile = try tmpFileWithContent("FOO=fromfile\nBAZ=fromfile\n")
+        defer { try? FileManager.default.removeItem(at: tmpFile) }
+
+        let result = try Parser.allEnv(
+            imageEnvs: ["FOO=fromimage", "BAR=fromimage"],
+            envFiles: [tmpFile.path],
+            envs: ["FOO=fromuser"]
+        )
+        #expect(Set(result) == Set(["FOO=fromuser", "BAR=fromimage", "BAZ=fromfile"]))
+    }
+
     private func tmpFileWithContent(_ content: String) throws -> URL {
         let tempDir = FileManager.default.temporaryDirectory
         let tempFile = tempDir.appendingPathComponent("envfile-test-\(UUID().uuidString)")
